#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 21 00:55:02 2021

@author: dboateng
"""

from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.preprocessing import StandardScaler, RobustScaler, Normalizer
from sklearn.preprocessing import QuantileTransformer, PowerTransformer
from sklearn.decomposition import PCA, KernelPCA, IncrementalPCA, SparsePCA
from copy import deepcopy
import numpy as np
import pandas as pd
import scipy.stats as st


def add_seasonal_cycle(t, anomalies, mean):

    """
    Adds a seasonal cycle such that

        X = anomalies + mean_seasonal_cycle

    Parameters
    ----------
    t : numpy array of ints
        time in number of months
    res : numpy array
        Array of standardized values
    mean : array of shape 12 x #columns(res)
        Mean values for each month and each column in res

    Returns
    -------
    X : unstandardized values
    """
    N = len(t)
    M = anomalies.shape[1]
    for i in range(N):
        for j in range(M):
            anomalies[i,j] += mean[t[i] % 12,j]
    return anomalies


def remove_seasonal_cycle(t, X, mean):

    """
    Inverse operation to add_seasonal_cycle
    """
    N = len(t)
    M = X.shape[1]
    for i in range(N):
        for j in range(M):
            X[i,j] -= mean[t[i] % 12,j]
    return X


def get_annual_mean_cycle(t, X):
    np.seterr("raise")
    
    N = len(t)
    M = X.shape[1]
    mean = np.zeros((12, M))
    counts = np.zeros((12, M), dtype=np.int)
    for i in range(N):
        for j in range(M):
            if not np.isnan(X[i,j]):
                mean[t[i] % 12,j] += X[i,j]
                counts[t[i] % 12,j] += 1
    for i in range(12):
        for j in range(M):
            mean[i,j] = mean[i,j]/counts[i,j]
    return mean


def get_mean_prediction(t, mean):
    N = len(t)
    mean_prediction = np.zeros(N)
    for i in range(N):
        mean_prediction[i] = mean[t[i] % 12, 0]
    return mean_prediction

class NoStandardizer(BaseEstimator, TransformerMixin):
    """
    This is just a dummy standardizer that does nothing.
    """

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X

    def inverse_transform(self, X):
        return X


class MonthlyStandardizer(BaseEstimator, TransformerMixin):
    """
    Standardizes monthly data that has a seasonal cycle and possibly a linear trend.

    Since the seasonal cycle might affect the trend estimation, the
    seasonal cycle is removed first (by subtracting the mean annual cycle)
    and the trend is estimated by linear regression. Afterwards the data is
    scaled to variance 1.

    Parameters
    ----------
    detrending : bool, optional (default: False)
        Whether to remove a linear trend
    """

    def __init__(self, detrending=False, scaling=False):
        self.detrending = detrending
        self.scaling = scaling

    def fit(self, X, y=None):
        """
        Fits the standardizer to the provided data, i.e. calculates annual mean
        cycle and trends.

        Parameters
        ----------
        X : pd.DataFrame or pd.Series
            DataFrame or Series which holds the data
        y : dummy (optional, default: None)
            Not used

        Returns
        -------
        self
        """
        # be careful: X might contain NaNs
        
        t = X.index.values.astype('datetime64[M]').astype(int)
        
        
        # if np.issubdtype(X.index.values.dtype, np.datetime64) == False:
        #     t = X.index.values.astype('datetime64[M]').astype(int)
            
        # else:
        #     t = X.index.values.astype(int)
            
        values = np.array(X.values) # deep copy
        # X might be a Series or a DataFrame
        if values.ndim == 1:
            values = values[:,np.newaxis]
        N, M = values.shape

        # get seasonal mean
        self.mean = get_annual_mean_cycle(t, values)


        # get slope
        if self.detrending:
            # remove seasonal cycle
            values = remove_seasonal_cycle(t, values, self.mean)

            self.slopes = np.zeros(M)
            self.intercepts = np.zeros(M)
            for col in range(M):
                # remove NaNs
                nans = np.isnan(values[:,col])
                x = values[~nans,col]
                time = t[~nans]
                slope, intercept, r, p, stderr = st.linregress(time, x)
                self.slopes[col] = slope
                self.intercepts[col] = intercept
            # remove trend
            values -= self.intercepts + np.outer(t, self.slopes)

        if self.scaling:
            self.std = np.nanstd(values, axis=0)

        return self

    def transform(self, X, y=None):
        """
        Standardizes the values based on the previously calculated parameters
        Parameters
        ----------
        X : pd.DataFrame or pd.Series
            DataFrame or Series which holds the data
        y : dummy (optional, default: None)
            Not used

        Returns
        -------
        X_transformed : pd.DataFrame or pd.Series
            Transformed data
        """
        t = X.index.values.astype('datetime64[M]').astype(int)
        values = np.array(X.values) # deep copy
        # X might be a Series or a DataFrame
        if values.ndim == 1:
            values = values[:,np.newaxis]

        values = remove_seasonal_cycle(t, values, self.mean)
        if self.detrending:
            values -= self.intercepts + np.outer(t, self.slopes) # remove trend

        if self.scaling:
            values /= self.std

        if X.values.ndim == 1:
            return pd.Series(data=values[:,0], index=X.index, name=X.name)
        else:
            return pd.DataFrame(data=values, index=X.index, columns=X.columns)


    def inverse_transform(self, X, y=None):
        """
        De-standardizes the values based on the previously calculated parameters
        Parameters
        ----------
        X : pd.DataFrame or pd.Series
            DataFrame or Series which holds the standardized data
        y : dummy (optional, default: None)
            Not used

        Returns
        -------
        X_unstandardized : pd.DataFrame or pd.Series
            Unstandardized data
        """
        t = X.index.values.astype('datetime64[M]').astype(int)
        values = np.array(X.values)
        # X might be a Series or a DataFrame
        if values.ndim == 1:
            values = values[:,np.newaxis]

        if self.scaling:
            values *= self.std

        # add the trend
        if self.detrending:
            values += self.intercepts + np.outer(t, self.slopes)

        # add seasonal cycle
        values = add_seasonal_cycle(t, values, self.mean)


        if X.values.ndim == 1:
            return pd.Series(data=values[:,0], index=X.index, name=X.name)
        else:
            return pd.DataFrame(data=values, index=X.index, columns=X.columns)


class StandardScaling(BaseEstimator, TransformerMixin):
    
    def __init__(self, method=None, with_std=True, with_mean=True, unit_variance=False,
                 norm="l2"):
        self.method = method 
        self.with_std = with_std
        self.with_mean = with_mean
        self.unit_variance = unit_variance
        self.norm = norm
        
        
        if self.method == None or "standardscaler":
            self.scaler = StandardScaler(with_mean=self.with_mean,
                                                with_std=self.with_std)
        elif self.method == "robustscaler":
            self.scaler = RobustScaler(with_centering=self.with_mean, with_scaling=self.with_std)
        
        elif self.method == "normalize":
            self.scaler = Normalizer(norm=self.norm)
        
        elif self.method == "powertransformer":
            self.scaler = PowerTransformer(method="yeo-johnson", standardize=True)
            
        elif self.method == "quantiletransformer":
            self.scaler = QuantileTransformer(n_quantiles=1000, output_distribution="unifom")
        
        else:
            raise ValueError("The standardizer do not recognize the defined method")
                
        
    def fit(self, X):
        X_values = X.values 
        if X_values.ndim == 1:
            X_values = X_values[:,np.newaxis]
        
        self.scaler.fit(X_values)
        
        return self 
    
    # def fit_transform(self, X, y=None):
        
    #     values = self.scaler.fit_transform(X=X, y=y)
        
    #     if X.values.ndim == 1:
    #         return pd.Series(data=values[:,0], index=X.index, name=X.name)
    #     else:
    #         return pd.DataFrame(data=values, index=X.index, columns=X.columns)
        
    
    def inverse_transform(self, X):
        
        X_values = X.values 
        if X_values.ndim == 1:
            X_values = X_values[:,np.newaxis]
        
        values = self.scaler.inverse_transform(X_values)
        
        if X.values.ndim == 1:
            return pd.Series(data=values[:,0], index=X.index, name=X.name)
        else:
            return pd.DataFrame(data=values, index=X.index, columns=X.columns)
        
    
    def transform(self, X):
        X_values = X.values 
        if X_values.ndim == 1:
            X_values = X_values[:,np.newaxis]
            
        values = self.scaler.transform(X_values)
        
        if X.values.ndim == 1:
            return pd.Series(data=values[:,0], index=X.index, name=X.name)
        else:
            return pd.DataFrame(data=values, index=X.index, columns=X.columns)
        
    
        
class PCAScaling(TransformerMixin, BaseEstimator):
    
    def __init__(self, n_components=None, kernel="linear", method=None):
        self.n_components = n_components
        self.kernel = kernel
        self.method = method
        
        
        if self.method == "PCA" or None:
            self.scaler = PCA(n_components=self.n_components)
            
        elif self.method == "IncrementalPCA":
            self.scaler = IncrementalPCA(n_components=self.n_components)
            
        elif self.method == "KernelPCA":
            self.scaler = KernelPCA(n_components=self.n_components, kernel=self.kernel)
            
        elif self.method == "SparsePCA":
            self.scaler = SparsePCA(n_components=self.n_components)
        
        else:
            raise ValueError("The method passed to the PCAscaling object is not defined")
            
            
    def fit(self, X):
        
        X_values = X.values 
        if X_values.ndim == 1:
            X_values = X_values[:,np.newaxis]
        
        self.scaler.fit(X_values)
        
    
    def fit_transform(self, X):
        
        X_values = X.values 
        if X_values.ndim == 1:
            X_values = X_values[:,np.newaxis]
            
        
        values = self.scaler.fit_transform(X_values)
        
        if X.values.ndim == 1:
            return pd.Series(data=values[:,0], index=X.index)
        else:
            return pd.DataFrame(data=values, index=X.index)
        
    
    def inverse_transform(self, X):
        
        X_values = X.values 
        if X_values.ndim == 1:
            X_values = X_values[:,np.newaxis]
        
        values = self.scaler.inverse_transform(X_values)
        
        if X.values.ndim == 1:
            return pd.Series(data=values[:,0], index=X.index)
        else:
            return pd.DataFrame(data=values, index=X.index)
    
    def transform(self, X):
        X_values = X.values 
        
        if X_values.ndim == 1:
            X_values = X_values[:,np.newaxis]
            
        values = self.scaler.transform(X_values)
        
        if X.values.ndim == 1:
            return pd.Series(data=values[:,0], index=X.index)
        else:
            return pd.DataFrame(data=values, index=X.index)
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 12 14:01:43 2021


This routine handles the preprocessing of data downloaded directly from DWD. The default time series is monthly, others frequency must be pass to the function
1. Extracting only stations with required number of years 
2. Writing additional information into files (eg. station name, lat, lon and elevation), since its downloaded into a separate file using station codes
3. All utils function to read stations into pyESD Station operator class

Note: This routine is specifically designed for data downloded from DWD (otherwise please contact daniel.boateng@uni-tuebingen.de for assistance on other datasets)

@author: dboateng
"""

# importing packages
import os 
import pandas as pd
import numpy as np

#from StationOperator import StationOperator
# #local packages
try:
    from .StationOperator import StationOperator
except:
    from StationOperator import StationOperator
    



def read_weatherstationnames(path_to_data):
    """
    This function reads all the station names in the data directory

    Parameters
    ----------
    path_to_data : TYPE: str
        DESCRIPTION. The directory path to where all the station data are stored

    Returns
    -------
    namedict : TYPE: dict
        DESCRIPTION.

    """

    nr, name = np.loadtxt(os.path.join(path_to_data, 'stationnames.csv'),
                              delimiter=',', skiprows=1, usecols=(0,1),
                              dtype=str, unpack=True)

    nr = [int(i) for i in nr]
    namedict = dict(zip(nr, name))
    return namedict


def read_station_csv(filename, varname, return_all=False):
    """
    

    Parameters
    ----------
    filename : TYPE: str
        DESCRIPTION. Name of the station in path 
    varname : TYPE: str
        DESCRIPTION. The name of the varibale to downscale (eg. Precipitation, Temperature)

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    ws : TYPE
        DESCRIPTION.

    """
    
    # reading headers info with readline 
    with open(filename, "r") as f:
        name = f.readline().split(',')[1].replace("\n", "")
        lat = float(f.readline().split(',')[1].replace("\n",""))
        lon = float(f.readline().split(',')[1].replace("\n",""))
        elev = float(f.readline().split(',')[1].replace("\n",""))
        
        
        
    data = pd.read_csv(filename, sep=',', skiprows=6, usecols=[0,1,],
                       parse_dates=[0], index_col=0, names=['Time',
                       varname])
    
    data = data.dropna()
    
    if varname == "Precipitation":
        pr = data[varname]
        pr[pr == -9999] = np.nan
        assert not np.any(pr < -1e-2)
        assert not np.any(pr > 2000)
        
        data = {varname:pr}
        
    elif varname == "Temperature":
        t = data[varname]
        t[t == -8888] = np.nan
        assert not np.any(t < -50)
        assert not np.any(t > 80)
        
        data = {varname:t}
        
    elif varname == "O18":
        d18O = data[varname]
        d18O[d18O == -9999] = np.nan
    else:
        raise ValueError("The model does not recognize the variable name")

    if return_all == False:

        so = StationOperator(data, name, lat, lon, elev)
        
        return so
    
    else:
        return data, lat, lon
    
    
    
    
def read_weatherstations(path_to_data):
    """
    Read all the station data in a directory.

    Parameters
    ----------
    path_to_data : TYPE: STR
        DESCRIPTION. relative or absolute path to the station folder

    Returns
    -------
    stations : TYPE: DICT
        DESCRIPTION. Dictionary containing all the datasets

    """
    namedict = read_weatherstationnames(path_to_data)
    stations = {}
    for i in namedict:
        filename = namedict[i].replace(' ', '_') + '.csv'
        print("Reading", filename)
        ws = read_station_csv(os.path.join(path_to_data, filename))
        stations[i] = ws
    return stations


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 26 11:15:31 2022

@author: dboateng
"""
# importing packages
import os 
import pandas as pd
import numpy as np
from pathlib import Path as p


def extract_DWDdata_with_more_yrs(path_to_data, path_to_store, min_yrs=60, glob_name="data*.csv", varname="Precipitation", start_date=None,
                          end_date=None, data_freq=None):
    """
    1. make directory to store the data that satisfies the time criteria
    2. Set a glob name that can be used to read all data in directory eg. *data*.csv or *.csv
    The function basically check if the station has more that certain number of year and its then written into another folder (path_to_store)

    Parameters
    ----------
    path_to_data : TYPE: str
        DESCRIPTION. path the directory containing all data 
    path_to_store : TYPE: str
        DESCRIPTION. path to store satisfied data
    yrs : TYPE: int
        DESCRIPTION. The number of years required for analysis
    glob_name : TYPE: str
        DESCRIPTION. The global pattern used to extract data (eg. *.csv)
    varname : TYPE: str, optional
        DESCRIPTION. The default is "Precipitation".
    start_date : TYPE:str, optional
        DESCRIPTION. The default is None. Start date required for the analysis eg. 1958-01-01
    end_date : TYPE: str, optional
        DESCRIPTION. The default is None. End date for data
    data_freq : TYPE:str, optional
        DESCRIPTION. The default is None. The frequency of data (eg. MS), must follow pandas setting

    Returns
    -------
    None.

    """
    if all(par is not None for par in [start_date, end_date, data_freq]):
        daterange = pd.date_range(start=start_date, end=end_date, freq=data_freq)
    else:
        print("------using the default time of ERA5 dataset: 1958-2020--------")
        
        daterange = pd.date_range(start="1958-01-01", end="2021-12-01", freq="MS") # time range for ERA5
    
    use_colums = ["Zeitstempel", "Wert"]
    
    df_to_store = pd.DataFrame(columns=["Time", varname,])
    
    df_to_store["Time"] = daterange
    df_to_store = df_to_store.set_index(["Time"], drop=True)
    
    
    # looping through directory
    for csv in p(path_to_data).glob(glob_name):
        df = pd.read_csv(csv, index_col=False, usecols=use_colums, parse_dates=["Zeitstempel"])
        df = df.set_index(["Zeitstempel"], drop=True)
        
        df = df.loc[daterange[0]:daterange[-1]] # selects only available dates
        
        if data_freq is None or data_freq =="MS":
            yrs = len(df)/12
        else:
            raise Exception("different time period is not implemented")
        
        if yrs >= min_yrs: # data with more than 60 years
        
        
            df_to_store[varname][df.index] = df["Wert"][df.index]
            
            if varname == "Precipitation":
                df_to_store = df_to_store.replace(np.nan, -9999)
                
                
            elif varname == "Temperature":
                df_to_store = df_to_store.replace(np.nan, -8888)
                
                
            else:
                raise ValueError("Incorrect variable name")
            
            print("saving", csv.name)
            
            df_to_store.to_csv(os.path.join(path_to_store, csv.name), index=True)
            
            
            
            
def add_info_to_data(path_to_info, path_to_data, path_to_store, glob_name="data*", 
                     varname="Precipitation"):
    """
    This function locate the data info in data_to_info by using the station code and then append it to the start of the data 
    This function also stores a summary info file (station names and station loc) that can be used to interate all the stations when applying the downscaling package

    Parameters
    ----------
    path_to_info : TYPE: str
        DESCRIPTION. Path to the data containing all the station infomation
    path_to_data : TYPE: str
        DESCRIPTION. Path to data required for appending info
    glob_name : TYPE: str
        DESCRIPTION. The global pattern used to extract data (eg. *.csv)

    Returns
    -------
    None.

    """
    info_cols = ["SDO_ID","SDO_Name","Geogr_Laenge","Geogr_Breite","Hoehe_ueber_NN"]
    
    
    df_info = pd.DataFrame(columns= info_cols[1:])
    
    for csv_file in p(path_to_data).glob(glob_name):
        sep_filename = csv_file.name.split(sep="_")
        csv_id = int(sep_filename[-1].split(sep=".")[0])
        
       # data = pd.read_csv(csv_file)
        data_info = pd.read_csv(path_to_info, usecols=info_cols)
        data_info = data_info.set_index(["SDO_ID"])
        
        df_info = df_info.append(data_info.loc[csv_id])
        
        csv_info = data_info.loc[csv_id]
        
        # reading data from path (csv_file)
        data_in_glob = pd.read_csv(csv_file)
        
        print(csv_file)
        
        time_len = len(data_in_glob["Time"])
        
        # creating new dataFrame to store the data in required format
        
        df = pd.DataFrame(index=np.arange(800), columns=np.arange(2))
        
        # filling the df up using csv_file
        df.at[0,0] = "Station"
        df.at[0,1] = csv_info["SDO_Name"]
        
        df.at[1,0] = "Latitude"
        df.at[1,1] = float(csv_info["Geogr_Breite"].replace(",", "."))
        
        df.at[2,0] = "Longitude"
        df.at[2,1] = float(csv_info["Geogr_Laenge"].replace(",", "."))
        
        df.at[3,0] = "Elevation"
        df.at[3,1] = csv_info["Hoehe_ueber_NN"]
        
        df.at[5,0] = "Time"
        df.at[5,1] = varname
        
        #adding data from sorted file
        for i in range(time_len):
            df.loc[6+i,0] = data_in_glob["Time"][i]
            df.loc[6+i,1] = data_in_glob[varname][i]
        
        # saving the file with station name
        
        name = csv_info["SDO_Name"]
        
        # replacing special characters eg. German umlaute
        special_char_map = {ord('ä'):'ae', ord('ü'):'ue', ord('ö'):'oe', ord('ß'):'ss', 
                            ord(",") : "", ord(" ") : "_", ord("/"):"_", ord("."):"",}
        name = name.translate(special_char_map)
        
        df.to_csv(os.path.join(path_to_store, name + ".csv"), index=False, header=False)
        
    df_info = df_info.rename(columns = {"SDO_Name":"Name","Geogr_Laenge":"Longitude",
                                        "Geogr_Breite":"Latitude","Hoehe_ueber_NN":"Elevation"})
    
    # replace comma with dot
    
    df_info["Latitude"] = df_info["Latitude"].apply(lambda x:x.replace(",","."))
    
    df_info["Longitude"] = df_info["Longitude"].apply(lambda x:x.replace(",","."))
    
    df_info["Name"] = df_info["Name"].apply(lambda x:x.translate(special_char_map))
    
    df_info = df_info.sort_values(by=["Name"], ascending=True)
    
    df_info = df_info.reset_index(drop=True)
                                 
    
    df_info.to_csv(os.path.join(path_to_store, "stationnames.csv"))
    
    df_info.drop("Name", axis=1, inplace=True)
    
    df_info.to_csv(os.path.join(path_to_store, "stationloc.csv"))
    
    
# add function for Ghana datsets



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan  3 17:18:14 2022

@author: dboateng
"""

# importing modules 
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.feature_selection import RFECV, SelectFromModel, SequentialFeatureSelector
from sklearn.linear_model import Lasso, LassoCV, Ridge, BayesianRidge, ARDRegression
from sklearn.inspection import permutation_importance
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import TimeSeriesSplit, LeaveOneOut, LeaveOneGroupOut
from sklearn.svm import SVR
from sklearn.ensemble import ExtraTreesRegressor, ExtraTreesClassifier

from pyESD.plot_utils import apply_style



# class for recursive feature elimination 
class RecursiveFeatureElimination():
    
    def __init__(self, regressor_name="ARD"):
        self.regressor_name = regressor_name
        self.cv = TimeSeriesSplit()
        self.n_jobs = -1
        self.min_features = 5
        self.scoring = "r2"
        
        
        if self.regressor_name == "ARD":
            self.estimator = ARDRegression()
        elif self.regressor_name == "BayesianRidge":
            self.estimator = BayesianRidge()
        elif self.regressor_name == "lasso":
            self.estimator = Lasso()
        elif self.regressor_name == "lassocv":
            self.estimator = LassoCV()
        elif self.regressor_name == "Ridge":
            self.estimator = Ridge()
        elif self.regressor_name == "RandomForest":
            self.estimator = RandomForestRegressor()
        else:
            raise ValueError("Check the regressor if implemented")
        
        
    def fit(self, X, y):
        
        self.regressor = RFECV(estimator=self.estimator, scoring=self.scoring, cv=self.cv, n_jobs= self.n_jobs,
                      min_features_to_select=self.min_features).fit(X, y)
        
    
    def print_selected_features(self, X):
        num_features = self.regressor.n_features_
        select_names = X.columns[self.regressor.support_]
        
        self.select_names = select_names
        
        print("{0} : optimal number of predictors and selected variables are {1}".format(num_features, select_names))
    
    def transform(self, X):
        X_new = self.regressor.transform(X)
        return X_new
    
    def cv_test_score(self):
        cv_score = self.regressor.cv_results_["mean_test_score"].mean()
        return cv_score
    
    def score(self,X,y):
        score = self.regressor.score(X,y)
        return score
        
        

class TreeBasedSelection():
    def __init__(self, regressor_name="RandomForest"):
        self.regressor_name = regressor_name
        self.n_jobs = -1
        self.bootstrap = True
        self.criterion = "squared_error"
        self.scoring = "r2"
        
        if self.regressor_name == "RandomForest":
            self.estimator = RandomForestRegressor(n_jobs=self.n_jobs, criterion=self.criterion, bootstrap=self.bootstrap, 
                                                   n_estimators=200)
        elif self.regressor_name == "ExtraTree":
            self.estimator = ExtraTreesRegressor(n_jobs=self.n_jobs, criterion=self.criterion, bootstrap=self.bootstrap)
        else:
            raise ValueError("Tree regressor estimator is not defined properly")
            
        
    def fit(self, X, y):
        
        self.regressor = SelectFromModel(estimator=self.estimator, prefit=False).fit(X,y)
        
    def transform(self, X):
        X_new = self.regressor.transform(X)
        return X_new
    
    def feature_importance(self, X,y, plot=False, fig_path=None, fig_name=None, save_fig=False, station_name=None):
        self.estimator.fit(X,y)
        importance = self.estimator.feature_importances_
        feature_names = X.columns
        forest_importances = pd.Series(importance, index=feature_names)
        
        if plot == True:
            apply_style(fontsize=28, style="seaborn-talk", linewidth=3, usetex=False)  
            std = np.std([tree.feature_importances_ for tree in self.estimator.estimators_], axis=0)
            fig,ax = plt.subplots(figsize=(15, 13))
            forest_importances.plot.bar(yerr=std, ax=ax)
            if station_name is not None:
                ax.set_title("Feature importances using tree regressor (" + station_name + ")",
                             fontweight="bold", fontsize=24)
                
            else:
                ax.set_title("Feature importances using tree regressor")
                
            ax.set_ylabel("Mean Decrease in impurity", fontweight="bold", fontsize=24)
            fig.tight_layout()
            if save_fig:
                plt.savefig(os.path.join(fig_path, fig_name), bbox_inches="tight", format= "png")
                
            else:
                plt.show()
            
        return forest_importances
    
    def permutation_importance_(self, X,y, plot=False, fig_path=None, fig_name=None, save_fig=False):
        self.estimator.fit(X,y)
        importance = permutation_importance(estimator=self.estimator, X=X, y=y, scoring=self.scoring,
                                            n_repeats=10, n_jobs=self.n_jobs)
        sorted_idx = importance.importance_mean.argsort()
        if plot == True:
            fig,ax = plt.subplots(figsize=(15, 13))
            ax.boxplot(importance.importances[sorted_idx].T, vert=False, labels=X.columns[sorted_idx])
            ax.set_title("Permutation Importances (On test data)")
            ax.set_ylabel("Mean accuracy decrease", fontweight="bold", fontsize=20)
            fig.tight_layout()
            if save_fig:
                plt.savefig(os.path.join(fig_path, fig_name), bbox_inches="tight", format= "png")
                
            else:
                plt.show()
            
        return sorted_idx
            
    
    def print_selected_features(self, X):
        select_names = X.columns[self.regressor.get_support(indices=True)]
        num_features = len(select_names)
        print("{0} : optimal number of predictors and selected variables are {1}".format(num_features, select_names))
        
        self.select_names = select_names
        

class SequentialFeatureSelection():
    def __init__(self, regressor_name = "Ridge", n_features= 10, direction="forward"):
        self.regressor_name = regressor_name
        self.n_features =n_features
        self.scoring = "r2"
        self.direction = direction
        
        if self.regressor_name == "ARD":
            self.estimator = ARDRegression()
        elif self.regressor_name == "BayesianRidge":
            self.estimator = BayesianRidge()
        elif self.regressor_name == "lasso":
            self.estimator = Lasso()
        elif self.regressor_name == "lassocv":
            self.estimator = LassoCV()
        elif self.regressor_name == "Ridge":
            self.estimator = Ridge()
        else:
            raise ValueError("Check the regressor if implemented")
            
    def fit(self, X,y):
        self.regressor = SequentialFeatureSelector(estimator=self.estimator, n_features_to_select=self.n_features, scoring=self.scoring,
                                                   direction=self.direction).fit(X,y)
        
    def score(self,X,y):
        score = self.regressor.score(X,y)
        return score
    
    def transform(self, X):
        X_new = self.regressor.transform(X)
        return X_new  
    
    def print_selected_features(self, X):
        select_names = X.columns[self.regressor.get_support(indices=True)]
        num_features = len(select_names)
        print("{0} : optimal number of predictors and selected variables are {1}".format(num_features, select_names))
        
        
        self.select_names = select_names
    



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 16 12:26:01 2022

@author: dboateng

This module require further development of add deep learning models!
"""

# importing models 

import tensorflow as tf
import tensorflow.keras as keras 
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout


class DeepLearningRegressor():
    
    def __init__(self, method=None, optimizer="adam", loss="mean_squared_error", metrics=["RootMeanSquaredError"]):
        self.optimizer = optimizer 
        self.loss = loss
        self.metrics = metrics
        self.method = method      
        
        if self.method == None:
            print(".....Dense Neural Network is used as default regressor......")
            self.method = "Dense"
            
    def build_model(self):
        if self.method == "Dense":
            print("....using default design...edit the dense_model.py if the Package was installed with the edit option")
            
            self.model = Sequential()
            self.model.add(Dense(512, activation="relu", input_dim=13))
            self.model.add(Dense(256, activation="relu"))
            self.model.add(Dropout(0.2))
            self.model.add(Dense(64, activation="relu"))
            self.model.add(Dense(1))
            
        elif self.method == "LSTM":
            
            raise NotImplementedError("LSTM is not implemented yet, check for future version")
            
        elif self.method == "CNN":
            raise NotImplementedError("CNN is not implemented yet, check for future version")
            
        else:
            raise ValueError("Not recognized method")

    
    def plot_network():
        pass
    
    def compile_model(self):
        return self.model.compile(optimizer=self.optimizer, loss = self.loss, metrics=self.metrics)
        
    
    def convert_to_sklearn_regressor(self, epochs=1000, verbose=False):
        self.estimator = tf.keras.wrappers.scikit_learn.KerasRegressor(self.model, epochs= epochs, verbose= verbose)
        self.estimator._estimator_type = "regressor"
        
        return self.estimator
        
    
    def fit(self, X, y):
        
        return self.estimator.fit(X,y)
    
    def predict(self, X):
        yhat = self.estimator.predict(X)
        return yhat

    
    
        
        


#!/usr/bin/env python

"""The setup script."""

import io
from os import path as op
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()


with open("HISTORY.rst") as history_file:
    history = history_file.read()

here = op.abspath(op.dirname(__file__))

# get the dependencies and installs

requirements = ["cftime>=1.6.0",
    "eofs>=1.4",
    "geopandas>=0.12",
    "numpy>=1.21",
    "pandas>=1.3",
    "scikit-learn>=0.24",
    "scikit-optimize>=0.9.0",
    "scipy>=1.7",
    "seaborn>=0.11",
    "tensorflow>=2.8.0",
    "xarray>=2023.1",
    "xgboost>=1.5",
    "cycler>=0.10"
    ]
dev_requirements = []
with open("requirements_dev.txt") as dev:
    for dependency in dev.readlines():
        dev_requirements.append(dependency)

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]
KEYWORDS = "pyESD Empirical Statistical Downscaling"

setup(
    author="Daniel Boateng",
    author_email='dannboateng@gmail.com',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    description="Python Package for Empirical Statistical Downscaling. pyESD is under active development and all colaborators are welcomed. The purpose of the package is to downscale any climate variables e.g. precipitation and temperature using predictors from  reanalysis datasets (eg. ERA5) to point scale. pyESD adopts many ML and AL as the transfer function. ",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords=KEYWORDS,
    name='pyESD',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/Dan-Boat/PyESD',
    extras_require={"dev": dev_requirements},
    version='1.0.7',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
Created on Fri May 13 14:23:23 2022

@author: dboateng
"""
import unittest
import numpy as np 

from pyESD.standardizer import MonthlyStandardizer, PCAScaling, StandardScaling

from test_syn_data import generate_syn_data

X,y = generate_syn_data()


class Test_Standardizer(unittest.TestCase):
    
    @classmethod
    def setUp(self):
        self.ms = MonthlyStandardizer(detrending=False)
        self.pca = PCAScaling(method="PCA")
        self.ss = StandardScaling(method="standardscaler")
        
        
    def test_fit_transform_ms(self):
        X_st = self.ms.fit_transform(X)
        
        self.assertFalse(np.all(X_st == X), "the monthly standardizer is not working, check fit")
        
    
    def test_fit_transform_pca(self):
        X_st = self.pca.fit_transform(X)
        
        self.assertFalse(np.all(X_st == X), "the monthly standardizer is not working, check fit")
    
    def test_fit_transform_ss(self):
        X_st = self.ss.fit_transform(X)
        
        self.assertFalse(np.all(X_st == X), "the monthly standardizer is not working, check fit")
        
        
        
if __name__ == '__main__': 
    unittest.main()


# -*- coding: utf-8 -*-
"""
Created on Mon May 16 15:59:40 2022

@author: dboateng
"""

import xarray as xr
import pandas as pd 
import os 


path_to_save = "C:/Users/dboateng/Desktop/Datasets/CMIP5/Monthly/RCP85"

path = "C:/Users/dboateng/Desktop/Datasets/CMIP5/Monthly/preprosessed"


#data = xr.open_dataset(os.path.join(path, "z250_monthly.nc"))

# formating script for cmip dataset (start month, remove other coords, change coordinate name)

for file in os.listdir(path):
    
    print(file)

    varname = file.split("_")[0]
      
    data = xr.open_dataset(os.path.join(path, file))
    
    
    if 'lev' in data.coords and len(data.lev) == 1:
        data = data.drop('lev')
    
    
    data["time"] = data.time.values.astype("datetime64[M]")
    
    if hasattr(data, "plev") == True:
        data[varname] = data[varname][:, 0, :, :]
    
    data = data[varname]
    
    
    data.to_netcdf(os.path.join(path_to_save, file))
    
    



# -*- coding: utf-8 -*-
"""
Created on Fri May 13 14:22:11 2022

@author: dboateng
"""

import unittest
import numpy as np
from sklearn.utils.validation import check_is_fitted


from pyESD.ensemble_models import EnsembleRegressor
from pyESD.models import Regressors
from test_syn_data import generate_syn_data

X,y = generate_syn_data()


regressors = ["AdaBoost", "LassoLarsCV", "ARD"]
method = "Stacking"


estimators = []

for i in range(len(regressors)):
    
    regressor = Regressors(method= regressors[i], cv=5)
    regressor.set_model()
    estimators.append((regressors[i], regressor.estimator))
    


class TestEnsembleRegressor(unittest.TestCase):
    
    @classmethod
    def setUp(self):
        self.ensemble = EnsembleRegressor(estimators=estimators, cv=5, 
                                          method=method)
        
        self.ensemble.fit(X,y)
        
    def test_fit(self):
        
        self.assertIsNone(check_is_fitted(self.ensemble.ensemble, "estimators_"), 
                          "The ensemble regressor might not be fitted")
        
    
    def test_score(self):
        score = self.ensemble.score(X,y)
        
        self.assertGreaterEqual(score, 0.99,
                                "The model is not well calibrating, check parameters")
    
    def test_predict(self):
        yhat = self.ensemble.predict(X)
        
        self.assertGreaterEqual(np.corrcoef(y,yhat)[0,1], 0.99,
                               "The model is not well calibrating, check parameters")
        pass
    
    def test_cross_val_score(self):
        
        val_score = self.ensemble.cross_val_score(X, y)
        
        self.assertGreaterEqual(np.mean(val_score), 0.99,
                                "The model is not well calibrating, check parameters")
        
if __name__ == '__main__': 
    unittest.main()


# -*- coding: utf-8 -*-
"""
Created on Thu May 12 14:06:41 2022

@author: dboateng
"""

import unittest
import numpy as np
from sklearn.utils.validation import check_is_fitted


from pyESD.models import Regressors
from test_syn_data import generate_syn_data

X,y = generate_syn_data()


class TestRegressors(unittest.TestCase):
    
    
    @classmethod
    def setUpClass(self):
        self.regressor_lasso = Regressors(method="LassoLarsCV", cv=5)
        
        self.regressor_mlp = Regressors(method="MLPRegressor", cv=5, 
                                        hyper_method= "GridSearchCV")
    
    @classmethod
    def routine(self):
        self.regressor_lasso.set_model()
        self.regressor_lasso.fit(X,y)
        
        self.regressor_mlp.set_model()
        self.regressor_mlp.fit(X,y)
        
        
    def test_set_model(self):
        self.routine()
        
        self.assertTrue(hasattr(self.regressor_lasso, "estimator"),
                        "the model has not initialise the estimator")
        
        self.assertTrue(hasattr(self.regressor_mlp, "hyper"), 
                        "the set model has not set the hyperoptimize class")
        
    
    def test_fit(self):
        self.routine()
        
        self.assertIsNone(check_is_fitted(self.regressor_lasso.estimator, "coef_"),
                          "the regressor is not fitted or the set_model must be apply first")
        
        self.assertIsNone(check_is_fitted(self.regressor_mlp.estimator, "coefs_"),
                          "the regressor is not fitted or the set_model must be apply first")
    
    def test_score(self):
        self.routine()
        
        score = self.regressor_lasso.score(X,y)
        
        self.assertGreaterEqual(score, 0.99,
                                "The model is not well calibrating, check parameters")
    
        score = self.regressor_mlp.score(X,y)
        
        self.assertGreaterEqual(score, 0.99,
                                "The model is not well calibrating, check parameters")
        pass
    
    def test_predict(self):
        self.routine()
        
        yhat_lasso = self.regressor_lasso.predict(X)
        
        yhat_mlp = self.regressor_mlp.predict(X)
        
        self.assertGreaterEqual(np.corrcoef(y,yhat_lasso)[0,1], 0.99,
                                "The model is not well calibrating, check parameters")
        
        self.assertGreaterEqual(np.corrcoef(y,yhat_mlp)[0,1], 0.99,
                                "The model is not well calibrating, check parameters")
        
    
    def test_cross_val_score(self):
        self.routine()
        
        val_score_lasso = self.regressor_lasso.cross_val_score(X, y)
        
        self.assertGreaterEqual(np.mean(val_score_lasso), 0.99,
                                "The model is not well calibrating, check parameters")
        
        val_score_mlp = self.regressor_mlp.cross_val_score(X, y)
        
        self.assertGreaterEqual(np.mean(val_score_mlp), 0.99,
                                "The model is not well calibrating, check parameters")
        


if __name__ == '__main__': 
    unittest.main()